from typing import Any, Dict, Optional
from xml.etree import ElementTree

from requests import Response, Session
from saturnfs.errors import ExpiredSignature, SaturnError


class AWSPresignedClient:
    """
    Handles basic error checking, and detecting
    if a presigned URL has expired.
    """

    def __init__(self) -> None:
        self.session = Session()

    def get(
        self, url: str, headers: Optional[Dict[str, str]] = None, stream: bool = False
    ) -> Response:
        response = self.session.get(url, headers=headers, stream=stream)
        self.check_errors(response)
        return response

    def put(
        self,
        url: str,
        data: Optional[Any] = None,
        headers: Optional[Dict[str, str]] = None,
    ) -> Response:
        response = self.session.put(url, data, headers=headers)
        self.check_errors(response)
        return response

    def check_errors(self, response: Response):
        if not response.ok:
            if self.is_expired(response):
                raise ExpiredSignature()
            raise SaturnError(response.text, status=500)

    def is_expired(self, response: Response) -> bool:
        if response.status_code == 403:
            expires: Optional[str] = None
            try:
                root = ElementTree.fromstring(response.text)
                expires = root.findtext("./Expires")
            except Exception:
                pass
            return expires is not None
        return False

    def parse_etag(self, response: Response) -> str:
        etag: Optional[str] = None
        if "ETag" in response.headers:
            # upload_part returns etag in header
            etag = response.headers["ETag"]
        elif response.headers.get("Content-Type") == "application/xml":
            # upload_part_copy returns etag in XML response body
            root = ElementTree.fromstring(response.text)
            namespace = root.tag.split("}")[0].lstrip("{")
            etag = root.findtext(f"./{{{namespace}}}ETag")

        if not etag:
            raise SaturnError("Failed to parse etag from response", status=500)
        return etag

    def close(self):
        self.session.close()



class SaturnError(Exception):
    def __init__(self, message: str, status: int = 400) -> None:
        self.message = message
        self.status = status
        super().__init__(message)


class ExpiredSignature(SaturnError):
    def __init__(self) -> None:
        self.message = "Presigned URL has expired"
        super().__init__(self.message, status=401)


class PathErrors:
    INVALID_REMOTE_PATH = "Invalid remote path. Expected format is sfs://<org>/<identity>/..."
    INVALID_REMOTE_FILE = (
        "Invalid remote file path. Expected format is sfs://<org>/<identity>/<file_path>"
    )
    AT_LEAST_ONE_REMOTE_PATH = "Either source or destination must be a remote path"



import os

from saturnfs.utils import Units

SATURN_BASE_URL = os.environ["SATURN_BASE_URL"]
SATURN_TOKEN = os.environ["SATURN_TOKEN"]

S3_MIN_PART_SIZE = 5 * Units.MiB
S3_MAX_PART_SIZE = 5 * Units.GiB
S3_MAX_NUM_PARTS = 10000
OBJECT_STORAGE_MAX_LIST_COUNT = 1000

SATURNFS_PROTOCOL = "sfs"
SATURNFS_FILE_PREFIX = f"{SATURNFS_PROTOCOL}://"



from typing import Any, Dict

from requests import Session
from saturnfs.api.base import BaseAPI


class DownloadAPI(BaseAPI):
    endpoint = "/api/object_storage/download"

    @classmethod
    def get(
        cls,
        session: Session,
        data: Dict[str, Any],
    ) -> Dict[str, Any]:
        url = cls.make_url()
        response = session.post(url, json=data)
        cls.check_error(response, 200)
        return response.json()


class BulkDownloadAPI(BaseAPI):
    endpoint = "/api/object_storage/bulk_download"

    @classmethod
    def get(cls, session: Session, data: Dict[str, Any]) -> Dict[str, Any]:
        url = cls.make_url()
        response = session.post(url, json=data)
        cls.check_error(response, 200)
        return response.json()



import sys

import click
from saturnfs.api.base import SaturnError
from saturnfs.cli.commands import cli


def entrypoint():
    try:
        cli()
    except SaturnError as e:
        click.echo(f"Error: {e.message}")
        sys.exit(1)
    except FileNotFoundError as e:
        click.echo(f"FileNotFoundError: {e}")
        sys.exit(1)


if __name__ == "__main__":
    entrypoint()



from typing import Any, Dict

from requests import Session
from saturnfs.api.base import BaseAPI


class ListAPI(BaseAPI):
    endpoint = "/api/object_storage"

    @classmethod
    def get(
        cls,
        session: Session,
        **query_args: Any,
    ) -> Dict[str, Any]:
        url = cls.make_url(query_args=query_args)
        response = session.get(url)
        cls.check_error(response, 200)
        return response.json()


class SharedAPI(BaseAPI):
    endpoint = "/api/object_storage/shared"

    @classmethod
    def get(
        cls,
        session: Session,
        **query_args: Any,
    ) -> Dict[str, Any]:
        url = cls.make_url(query_args=query_args)
        response = session.get(url)
        cls.check_error(response, 200)
        return response.json()


class OrgListAPI(BaseAPI):
    endpoint = "/api/orgs"

    @classmethod
    def get(cls, session: Session) -> Dict[str, Any]:
        url = cls.make_url()
        response = session.get(url)
        cls.check_error(response, 200)
        return response.json()



from __future__ import annotations

from typing import List, Tuple, Union

import marshmallow_dataclass
from saturnfs import settings
from saturnfs.errors import PathErrors, SaturnError
from saturnfs.schemas.base import DataclassSchema


@marshmallow_dataclass.dataclass
class ObjectStorage(DataclassSchema):
    file_path: str
    owner_name: str

    @classmethod
    def parse(
        cls, remote_path: Union[str, ObjectStorage, ObjectStoragePrefix], **override: str
    ) -> ObjectStorage:
        owner_name, file_path = parse_remote(remote_path)
        data = {
            "owner_name": owner_name,
            "file_path": file_path,
            **override,
        }
        if not data.get("file_path"):
            raise SaturnError(PathErrors.INVALID_REMOTE_FILE)

        return cls(**data)

    @property
    def name(self) -> str:
        return full_path(self.owner_name, self.file_path)

    def __fspath__(self) -> str:
        return self.name

    def __str__(self) -> str:
        return f"{settings.SATURNFS_FILE_PREFIX}{self.name}"


@marshmallow_dataclass.dataclass
class ObjectStoragePrefix(DataclassSchema):
    prefix: str
    owner_name: str

    @classmethod
    def parse(
        cls, remote_prefix: Union[str, ObjectStorage, ObjectStoragePrefix], **override: str
    ) -> ObjectStoragePrefix:
        owner_name, prefix = parse_remote(remote_prefix)
        data = {
            "owner_name": owner_name,
            "prefix": prefix,
            **override,
        }
        return cls(**data)

    @property
    def name(self) -> str:
        return full_path(self.owner_name, self.prefix)

    def __fspath__(self) -> str:
        return self.name

    def __str__(self) -> str:
        return f"{settings.SATURNFS_FILE_PREFIX}{self.name}"


@marshmallow_dataclass.dataclass
class BulkObjectStorage(DataclassSchema):
    file_paths: List[str]
    owner_name: str


def parse_remote(path: Union[str, ObjectStorage, ObjectStoragePrefix]) -> Tuple[str, str]:
    """
    Convert any remote reference into a tuple of strings (<owner_name>, <path>)
    """
    if isinstance(path, str):
        if path.startswith(settings.SATURNFS_FILE_PREFIX):
            # Strip sfs://
            path = path[len(settings.SATURNFS_FILE_PREFIX) :]
        elif path.startswith("/"):
            # Allow /org/identity/...
            path = path[1:]

        path_split = path.split("/", 2)
        if len(path_split) < 2 or not all(path_split[:2]):
            raise SaturnError(PathErrors.INVALID_REMOTE_PATH)

        org_name = path_split[0]
        identity_name = path_split[1]
        owner_name = f"{org_name}/{identity_name}"
        if len(path_split) < 3:
            path = ""
        else:
            path = path_split[2]
    elif isinstance(path, ObjectStorage):
        owner_name = path.owner_name
        path = path.file_path
    else:
        owner_name = path.owner_name
        path = path.prefix

    return (owner_name, path)


def full_path(owner_name: str, path: str) -> str:
    return f"{owner_name.rstrip('/')}/{path.lstrip('/')}"



from enum import Enum


class Units(int, Enum):
    KiB = 1 << 10
    MiB = 1 << 20
    GiB = 1 << 30
    TiB = 1 << 40


def human_readable_format(size: float):
    for unit in ["B", "KiB", "MiB", "GiB", "TiB"]:
        if size < 1024.0 or unit == "TiB":
            break
        size /= 1024.0
    formatted = f"{size:.2f}"
    formatted = formatted.rstrip("0").rstrip(".")
    return f"{formatted} {unit}"



from __future__ import annotations

from typing import Dict, Optional

import click
from fsspec.callbacks import Callback
from saturnfs import settings


class FileOpCallback(Callback):
    """
    Simple callback to print file upload/download/copy/delete operations
    at the outer layer, followed by a "." for each chunk that is transferred.
    """

    def __init__(self, operation: str, inner: bool = False, first_branch: bool = False, **kwargs):
        self.operation = operation
        self.inner = inner
        self.first_branch = first_branch
        super().__init__(**kwargs)

    def branch(self, path_1: str, path_2: str, kwargs: Dict):
        if not self.first_branch:
            self.first_branch = True
        else:
            click.echo()

        description = self._description(path_1, path_2)
        click.echo(f"{self.operation}: {description} ", nl=False)
        callback = FileOpCallback(self.operation, inner=True)
        kwargs["callback"] = callback

    def call(self, hook_name: Optional[str] = None, **kwargs):
        if hook_name:
            super().call(hook_name=hook_name, **kwargs)
        elif self.inner:
            click.echo(".", nl=False)

    def set_size(self, size: int):
        self.size = size

    def _description(self, path_1: str, path_2: str) -> str:
        if self.operation in {"download", "copy"}:
            path_1 = self._remote_format(path_1)
        if self.operation in {"upload", "copy"}:
            path_2 = self._remote_format(path_2)
        if path_2:
            return f"{path_1} to {path_2}"
        return path_1

    def _remote_format(self, path: str) -> str:
        if path and not path.startswith(settings.SATURNFS_FILE_PREFIX):
            path = path.lstrip("/")
            path = f"{settings.SATURNFS_FILE_PREFIX}{path}"
        return path


def file_op(src_is_local: bool, dst_is_local: bool) -> str:
    if src_is_local:
        return "upload"
    elif dst_is_local:
        return "download"
    return "copy"



from json import JSONDecodeError
from typing import Any, Dict, Optional
from urllib.parse import urlencode, urljoin

from requests import Response
from saturnfs import settings
from saturnfs.errors import SaturnError


class BaseAPI:
    endpoint = "/"

    @classmethod
    def make_url(
        cls, subpath: Optional[str] = None, query_args: Optional[Dict[str, Any]] = None
    ) -> str:
        url = urljoin(settings.SATURN_BASE_URL, cls.endpoint)
        if subpath:
            subpath = subpath.lstrip("/")
            url = url.rstrip("/") + f"/{subpath}"
        if query_args:
            url += "?" + urlencode({k: v for k, v in query_args.items() if v is not None})
        return url

    @classmethod
    def check_error(cls, response: Response, expected_status: int) -> None:
        if response.status_code != expected_status:
            if response.content:
                try:
                    error = response.json()
                    raise SaturnError(error.get("message"), response.status_code)
                except JSONDecodeError:
                    if response.status_code == 404:
                        raise SaturnError(  # pylint: disable=raise-missing-from
                            "ObjectStorage is not enabled on this installation",
                            response.status_code,
                        )
                    raise SaturnError(  # pylint: disable=raise-missing-from
                        response.text, response.status_code
                    )



from typing import Any, Dict

from requests import Session
from saturnfs.api.base import BaseAPI


class UsageAPI(BaseAPI):
    endpoint = "/api/object_storage/usage"

    @classmethod
    def get(cls, session: Session, **query_args: Any) -> Dict[str, Any]:
        url = cls.make_url(query_args=query_args)
        response = session.get(url)
        cls.check_error(response, 200)
        return response.json()



from typing import Optional

import marshmallow_dataclass
from saturnfs.schemas.base import DataclassSchema


@marshmallow_dataclass.dataclass
class ObjectStorageUsageResults(DataclassSchema):
    used_bytes: int
    reserved_bytes: int
    file_count: int
    active_uploads: int

    max_bytes: Optional[int]
    max_files: Optional[int]
    max_uploads: int

    @property
    def remaining_bytes(self) -> Optional[int]:
        if self.max_bytes is not None:
            return self.max_bytes - self.used_bytes - self.reserved_bytes
        return None



import json
import os
import shutil
import sys
from glob import has_magic
from typing import Dict, Optional

import click
from fsspec import Callback
from fsspec.callbacks import NoOpCallback
from fsspec.implementations.local import make_path_posix
from saturnfs import settings
from saturnfs.cli.callback import FileOpCallback, file_op
from saturnfs.cli.utils import (
    OutputFormats,
    print_file_table,
    print_json,
    print_upload_table,
    strip_glob,
)
from saturnfs.client import SaturnFS
from saturnfs.errors import PathErrors, SaturnError


@click.group()
def cli():
    pass


@cli.command("cp")
@click.argument("source_path", type=str)
@click.argument("destination_path", type=str)
@click.option(
    "--part-size",
    "-p",
    type=int,
    default=None,
    help="Max part size in bytes for uploading or copying a file in chunks",
)
@click.option(
    "--recursive", "-r", is_flag=True, default=False, help="Copy files under a prefix recursively"
)
@click.option("--quiet", "-q", is_flag=True, default=False, help="Do not print file operations")
def copy(
    source_path: str, destination_path: str, part_size: Optional[int], recursive: bool, quiet: bool
):
    sfs = SaturnFS()
    src_is_local = not source_path.startswith(settings.SATURNFS_FILE_PREFIX)
    dst_is_local = not destination_path.startswith(settings.SATURNFS_FILE_PREFIX)

    if src_is_local and dst_is_local:
        raise SaturnError(PathErrors.AT_LEAST_ONE_REMOTE_PATH)

    if quiet:
        callback = NoOpCallback()
    else:
        operation = file_op(src_is_local, dst_is_local)
        callback = FileOpCallback(operation=operation)

    if src_is_local:
        sfs.put(
            source_path,
            destination_path,
            recursive=recursive,
            part_size=part_size,
            callback=callback,
        )
    elif dst_is_local:
        sfs.get(
            source_path,
            destination_path,
            recursive=recursive,
            callback=callback,
        )
    else:
        sfs.cp(
            source_path,
            destination_path,
            recursive=recursive,
            part_size=part_size,
            callback=callback,
        )

    if not quiet:
        click.echo()


@cli.command("mv")
@click.argument("source_path", type=str)
@click.argument("destination_path", type=str)
@click.option(
    "--part-size",
    "-p",
    type=int,
    default=None,
    help="Max part size in bytes for uploading or copying a file in chunks",
)
@click.option(
    "--recursive", "-r", is_flag=True, default=False, help="Copy files under a prefix recursively"
)
@click.option("--quiet", "-q", is_flag=True, default=False, help="Do not print file operations")
def move(
    source_path: str, destination_path: str, part_size: Optional[int], recursive: bool, quiet: bool
):
    sfs = SaturnFS()
    src_is_local = not source_path.startswith(settings.SATURNFS_FILE_PREFIX)
    dst_is_local = not destination_path.startswith(settings.SATURNFS_FILE_PREFIX)

    if src_is_local and dst_is_local:
        raise SaturnError(PathErrors.AT_LEAST_ONE_REMOTE_PATH)

    if quiet:
        callback = NoOpCallback()
        rm_callback = NoOpCallback()
    else:
        operation = file_op(src_is_local, dst_is_local)
        callback = FileOpCallback(operation=operation)
        rm_callback = FileOpCallback(operation="delete", inner=True, first_branch=True)

    if src_is_local:
        sfs.put(
            source_path,
            destination_path,
            part_size=part_size,
            recursive=recursive,
            callback=callback,
        )
        if isinstance(rm_callback, FileOpCallback):
            kwargs: Dict[str, Callback] = {}
            rm_callback.branch(make_path_posix(source_path), "", kwargs)
            rm_callback = kwargs["callback"]
            rm_callback.set_size(callback.size)
        if recursive:
            shutil.rmtree(source_path)
        else:
            os.remove(source_path)
        rm_callback.relative_update(callback.size)
    elif dst_is_local:
        sfs.get(
            source_path,
            destination_path,
            recursive=recursive,
            callback=callback,
        )
        if isinstance(rm_callback, FileOpCallback):
            kwargs: Dict[str, Callback] = {}  # type: ignore[no-redef]
            rm_callback.branch(source_path, "", kwargs)
        sfs.rm(source_path, recursive=recursive, **kwargs)
    else:
        sfs.mv(
            source_path,
            destination_path,
            part_size=part_size,
            recursive=recursive,
            callback=callback,
            rm_callback=rm_callback,
        )

    if not quiet:
        click.echo()


@cli.command("rm")
@click.argument("path", type=str)
@click.option(
    "--recursive",
    "-r",
    is_flag=True,
    default=False,
    help="Delete all files under a prefix recursively",
)
def delete(path: str, recursive: bool):
    sfs = SaturnFS()
    sfs.rm(path, recursive=recursive)


@cli.command("ls")
@click.argument("prefix", type=str, required=False, default="")
@click.option(
    "--recursive",
    "-r",
    is_flag=True,
    default=False,
    help="List all files recursively under the given prefix",
)
@click.option(
    "--output",
    "-o",
    default=OutputFormats.TABLE,
    help="Output format (table, json)",
)
@click.option(
    "--human-readable",
    "-h",
    is_flag=True,
    help="Display size in human readable units. Only used for table output.",
)
@click.option(
    "--absolute",
    "-a",
    is_flag=True,
    help="Display absolute path. Only used for table output.",
)
def ls(
    prefix: str,
    recursive: bool,
    output: str,
    human_readable: bool,
    absolute: bool,
):
    OutputFormats.validate(output)

    sfs = SaturnFS()
    if recursive:
        details = sfs.find(prefix, detail=True)
        results = list(details.values())  # pylint: disable=no-member
    elif has_magic(prefix):
        details = sfs.glob(prefix, detail=True)
        results = list(details.values())  # pylint: disable=no-member
    else:
        results = sfs.ls(prefix, detail=True)

    if output == OutputFormats.JSON:
        print_json([info.dump_extended() for info in results])
    elif output == OutputFormats.TABLE:
        relative_prefix = None
        if not absolute:
            relative_prefix = sfs._strip_protocol(prefix)
            relative_prefix = strip_glob(relative_prefix)
        print_file_table(results, human_readable=human_readable, relative_prefix=relative_prefix)


@cli.command("list-uploads")
@click.argument("prefix", type=str)
@click.option("--is-copy", type=bool, is_flag=True, help="List uploads with a copy source")
@click.option("--is-not-copy", type=bool, is_flag=True, help="List uploads with no copy source")
@click.option(
    "--output",
    "-o",
    default=OutputFormats.TABLE,
    help="Output format (table, json)",
)
@click.option(
    "--human-readable",
    "-h",
    is_flag=True,
    help="Display size in human readable units",
)
def list_uploads(
    prefix: str, is_copy: Optional[bool], is_not_copy: bool, output: str, human_readable: bool
):
    OutputFormats.validate(output)

    sfs = SaturnFS()
    if is_copy and is_not_copy:
        uploads = []
    else:
        if not is_copy:
            is_copy = False if is_not_copy else None
        uploads = sfs.list_uploads(prefix, is_copy=is_copy)

    if output == OutputFormats.JSON:
        print_json([upload.dump_extended() for upload in uploads])
    elif output == OutputFormats.TABLE:
        print_upload_table(uploads, is_not_copy=is_not_copy, human_readable=human_readable)


@cli.command("cancel-upload")
@click.argument("upload_id", type=str)
def cancel_upload(upload_id: str):
    sfs = SaturnFS()
    sfs.cancel_upload(upload_id)


@cli.command("exists")
@click.argument("path", type=str)
def exists(path: str):
    sfs = SaturnFS()
    path_exists = sfs.exists(path)
    click.echo(path_exists)
    if not path_exists:
        sys.exit(1)


@cli.command("usage")
@click.option("--owner", type=str, help="Owner name '<org>/<identity>'")
def storage_usage(owner_name: Optional[str] = None):
    sfs = SaturnFS()
    usage = sfs.usage(owner_name)
    click.echo(json.dumps(usage.dump(), indent=2))



from __future__ import annotations

from dataclasses import field
from datetime import datetime
from typing import Any, Dict, List

import marshmallow_dataclass
from saturnfs.schemas.base import DataclassSchema
from saturnfs.schemas.reference import full_path


@marshmallow_dataclass.dataclass
class ObjectStorageBulkDownload(DataclassSchema):
    files: List[ObjectStoragePresignedDownload] = field(default_factory=list)

    @classmethod
    def load_extended(
        cls, data: Dict[str, List[Dict[str, Any]]], owner_name: str
    ) -> ObjectStorageBulkDownload:
        for file in data.get("files", {}):
            file.setdefault("owner_name", owner_name)
        return cls.load(data)


@marshmallow_dataclass.dataclass
class ObjectStoragePresignedDownload(DataclassSchema):
    file_path: str
    updated_at: datetime
    url: str

    # Not returned from API, added during load
    owner_name: str

    @property
    def name(self) -> str:
        return full_path(self.owner_name, self.file_path)



import json
from typing import Any, ClassVar, Dict, List, Optional, Type, TypeVar, Union

from marshmallow import EXCLUDE
from marshmallow import Schema as BaseSchema
from saturnfs.errors import SaturnError

# Stand-in for https://peps.python.org/pep-0673/
Self = TypeVar("Self", bound="DataclassSchema")


class DataclassSchema:
    """
    Typing and other useful helpers for classes using @marshmallow_dataclass.dataclass
    """

    # For mypy. marshmallow_dataclass will replace this with the real schema
    Schema: ClassVar[Type[BaseSchema]]

    class Meta:
        ordered = True
        unknown = EXCLUDE

    @classmethod
    def load(cls: Type[Self], data: Dict[str, Any], **kwargs) -> Self:
        # Exclude unknowns so old client version doesn't break if new data is added to schema
        return cls.Schema().load(data, **kwargs)

    @classmethod
    def loads(cls: Type[Self], data: Union[str, bytes], **kwargs) -> Self:
        try:
            return cls.load(json.loads(data), **kwargs)
        except json.JSONDecodeError as e:
            raise SaturnError("Invalid JSON body") from e

    def dump(self, only: Optional[List[str]] = None) -> Dict[str, Any]:
        return self.Schema(only=only).dump(self)

    def __getitem__(self, key):
        return getattr(self, key)



from saturnfs.schemas.delete import ObjectStorageBulkDeleteResults  # noqa
from saturnfs.schemas.download import ObjectStorageBulkDownload  # noqa
from saturnfs.schemas.download import ObjectStoragePresignedDownload
from saturnfs.schemas.list import ObjectStorageInfo  # noqa
from saturnfs.schemas.reference import ObjectStorage, ObjectStoragePrefix  # noqa
from saturnfs.schemas.upload import ObjectStorageCompletedUpload  # noqa
from saturnfs.schemas.upload import (  # noqa
    ObjectStorageCompletePart,
    ObjectStoragePresignedPart,
    ObjectStoragePresignedUpload,
)



import os
from datetime import datetime
from typing import Any, BinaryIO, List, Optional, Tuple

from fsspec import Callback
from saturnfs.client.aws import AWSPresignedClient
from saturnfs.errors import ExpiredSignature
from saturnfs.schemas import (
    ObjectStorageCompletePart,
    ObjectStoragePresignedDownload,
    ObjectStoragePresignedUpload,
)
from saturnfs.schemas.upload import ObjectStoragePresignedPart


class FileTransferClient:
    """
    Translates copy commands to specific upload/download/copy operations
    """

    def __init__(self):
        self.aws = AWSPresignedClient()

    def upload(
        self, local_path: str, presigned_upload: ObjectStoragePresignedUpload, file_offset: int = 0
    ) -> Tuple[List[ObjectStorageCompletePart], bool]:
        completed_parts: List[ObjectStorageCompletePart] = []
        with open(local_path, "rb") as f:
            f.seek(file_offset)
            for part in presigned_upload.parts:
                chunk = FileLimiter(f, part.size)
                try:
                    completed_parts.append(self.upload_part(chunk, part))
                except ExpiredSignature:
                    return completed_parts, False
        return completed_parts, True

    def upload_part(self, data: Any, part: ObjectStoragePresignedPart) -> ObjectStorageCompletePart:
        response = self.aws.put(
            part.url,
            data,
            headers={
                "Content-Type": "application/octet-stream",
                "Content-Length": str(part.size),
                **part.headers,
            },
        )
        etag = self.aws.parse_etag(response)
        return ObjectStorageCompletePart(part_number=part.part_number, etag=etag)

    def copy(
        self, presigned_copy: ObjectStoragePresignedUpload
    ) -> Tuple[List[ObjectStorageCompletePart], bool]:
        completed_parts: List[ObjectStorageCompletePart] = []
        for part in presigned_copy.parts:
            try:
                completed_part = self.copy_part(part)
                completed_parts.append(completed_part)
            except ExpiredSignature:
                return completed_parts, False
        return completed_parts, True

    def copy_part(self, copy_part: ObjectStoragePresignedPart) -> ObjectStorageCompletePart:
        response = self.aws.put(copy_part.url, headers=copy_part.headers)
        etag = self.aws.parse_etag(response)
        return ObjectStorageCompletePart(part_number=copy_part.part_number, etag=etag)

    def download(
        self,
        presigned_download: ObjectStoragePresignedDownload,
        local_path: str,
        callback: Optional[Callback] = None,
    ):
        dirname = os.path.dirname(local_path)
        if dirname:
            os.makedirs(dirname, exist_ok=True)

        response = self.aws.get(presigned_download.url, stream=True)
        if callback is not None:
            content_length = response.headers.get("Content-Length")
            callback.set_size(int(content_length) if content_length else None)

        with open(local_path, "wb") as f:
            for chunk in response.iter_content(None):
                chunk_size = f.write(chunk)
                if callback is not None:
                    callback.relative_update(chunk_size)

        if callback is not None and callback.size == 0:
            callback.relative_update(0)

        set_last_modified(local_path, presigned_download.updated_at)

    def close(self):
        self.aws.close()


class FileLimiter:
    """
    File-like object that limits the max number of bytes to be
    read from the current position of an open file
    """

    def __init__(self, file: BinaryIO, max_bytes: int):
        self.file = file
        self.max_bytes = max_bytes
        self.bytes_read = 0

        # Ensures requests will stream this instead of attempting
        # to treat it as iterable chunks
        self.len = max_bytes

    def read(self, amount: int = -1) -> bytes:
        if self.bytes_read >= self.max_bytes:
            return b""

        bytes_remaining = self.max_bytes - self.bytes_read
        data = self.file.read(min(amount, bytes_remaining))
        self.bytes_read += len(data)
        return data


def set_last_modified(local_path: str, last_modified: datetime):
    timestamp = last_modified.timestamp()
    os.utime(local_path, (timestamp, timestamp))



from __future__ import annotations

from dataclasses import field
from datetime import datetime
from typing import Any, Dict, List, Optional

import marshmallow_dataclass
from saturnfs import settings
from saturnfs.schemas.base import DataclassSchema
from saturnfs.schemas.reference import ObjectStoragePrefix, full_path
from typing_extensions import Literal


@marshmallow_dataclass.dataclass
class ObjectStorageListResult(DataclassSchema):
    dirs: List[ObjectStorageInfo] = field(default_factory=list)
    files: List[ObjectStorageInfo] = field(default_factory=list)
    next_last_key: Optional[str] = None

    @classmethod
    def load_extended(
        cls, data: Dict[str, Any], prefix: ObjectStoragePrefix
    ) -> ObjectStorageListResult:
        dirs = data.pop("dirs", [])
        files = data.pop("files", [])
        data["dirs"] = [None] * len(dirs)
        data["files"] = [None] * len(files)

        for i, dir in enumerate(dirs):
            dir_details = ObjectStorageDirDetails.load(dir)
            data["dirs"][i] = {
                "file_path": dir_details.prefix,
                "owner_name": prefix.owner_name,
                "type": "directory",
            }
        for i, file in enumerate(files):
            file_details = ObjectStorageFileDetails.load(file)
            data["files"][i] = {
                **file_details.dump(),
                "owner_name": prefix.owner_name,
                "type": "file",
            }

        return cls.load(data)


@marshmallow_dataclass.dataclass
class ObjectStorageInfo(DataclassSchema):
    """
    Describes info about a file or directory in object storage

    Unified view of ObjectStorageFileDetails and ObjectStorageDirDetails for fsspec
    """

    file_path: str
    owner_name: str
    type: Literal["directory", "file"]
    size: int = field(default=0)
    created_at: Optional[datetime] = field(default=None)
    updated_at: Optional[datetime] = field(default=None)

    @property
    def is_dir(self) -> bool:
        return self.type == "directory"

    @property
    def name(self) -> str:
        return full_path(self.owner_name, self.file_path)

    def __fspath__(self) -> str:
        return self.name

    def __str__(self) -> str:
        return f"{settings.SATURNFS_FILE_PREFIX}{self.name}"

    def dump_extended(self) -> Dict[str, Any]:
        data = self.dump()
        data["name"] = self.name
        return data


@marshmallow_dataclass.dataclass
class ObjectStorageFileDetails(DataclassSchema):
    file_path: str
    size: int
    created_at: datetime
    updated_at: datetime


@marshmallow_dataclass.dataclass
class ObjectStorageDirDetails(DataclassSchema):
    prefix: str


@marshmallow_dataclass.dataclass
class ObjectStorageSharedResult(DataclassSchema):
    owners: List[ObjectStorageSharedOwner]
    next_last_key: Optional[str] = None


@marshmallow_dataclass.dataclass
class ObjectStorageSharedOwner(DataclassSchema):
    id: str
    name: str
    count: int


@marshmallow_dataclass.dataclass
class ObjectStorageOrgs(DataclassSchema):
    orgs: List[Org]


@marshmallow_dataclass.dataclass
class Org(DataclassSchema):
    id: str
    created_at: datetime
    name: str
    email: Optional[str]
    description: str
    is_default: bool
    locked: bool



from __future__ import annotations

from dataclasses import field
from typing import List

import marshmallow_dataclass
from saturnfs.schemas.base import DataclassSchema


@marshmallow_dataclass.dataclass
class ObjectStorageBulkDeleteResults(DataclassSchema):
    failed_file_paths: List[str] = field(default_factory=list)



from __future__ import annotations

from dataclasses import field
from datetime import datetime
from typing import Any, Dict, List, Optional

import marshmallow_dataclass
from saturnfs.schemas.base import DataclassSchema
from saturnfs.schemas.reference import ObjectStorage, ObjectStoragePrefix, full_path


@marshmallow_dataclass.dataclass
class ObjectStoragePresignedUpload(DataclassSchema):
    object_storage_upload_id: str
    is_copy: bool = False
    parts: List[ObjectStoragePresignedPart] = field(default_factory=list)

    @property
    def upload_id(self) -> str:
        return self.object_storage_upload_id


@marshmallow_dataclass.dataclass
class ObjectStoragePresignedPart(DataclassSchema):
    url: str
    part_number: int
    size: int
    headers: Dict[str, str] = field(default_factory=dict)


@marshmallow_dataclass.dataclass
class ObjectStorageCompletedUpload(DataclassSchema):
    parts: List[ObjectStorageCompletePart]


@marshmallow_dataclass.dataclass
class ObjectStorageCompletePart(DataclassSchema):
    etag: str
    part_number: int


@marshmallow_dataclass.dataclass
class ObjectStorageUploadList(DataclassSchema):
    uploads: List[ObjectStorageUploadInfo]

    @classmethod
    def load_extended(
        cls, data: Dict[str, Any], prefix: ObjectStoragePrefix
    ) -> ObjectStorageUploadList:
        uploads = data.get("uploads", [])
        for upload in uploads:
            upload["owner_name"] = prefix.owner_name
        return cls.load(data)


@marshmallow_dataclass.dataclass
class ObjectStorageUploadInfo(DataclassSchema):
    id: str
    file_path: str
    size: Optional[int]
    part_size: int
    expires_at: datetime
    copy_source: Optional[ObjectStorage]

    # Not returned from API, added during load
    owner_name: str = field()

    @property
    def name(self) -> str:
        return full_path(self.owner_name, self.file_path)

    def dump_extended(self) -> Dict[str, Any]:
        data = self.dump()
        data["name"] = self.name
        return data



from __future__ import annotations

import os
import weakref
from datetime import datetime
from glob import has_magic
from io import BytesIO, TextIOWrapper
from typing import Any, Dict, Iterable, List, Optional, Tuple, Union, overload
from urllib.parse import urlparse

from fsspec.callbacks import Callback, NoOpCallback
from fsspec.implementations.local import LocalFileSystem, make_path_posix
from fsspec.spec import AbstractBufferedFile, AbstractFileSystem
from fsspec.utils import other_paths
from saturnfs import settings
from saturnfs.client.file_transfer import FileTransferClient
from saturnfs.client.object_storage import ObjectStorageClient
from saturnfs.errors import ExpiredSignature, SaturnError
from saturnfs.schemas import ObjectStorage, ObjectStoragePrefix
from saturnfs.schemas.download import ObjectStoragePresignedDownload
from saturnfs.schemas.list import ObjectStorageInfo
from saturnfs.schemas.reference import BulkObjectStorage
from saturnfs.schemas.upload import (
    ObjectStorageCompletePart,
    ObjectStoragePresignedPart,
    ObjectStoragePresignedUpload,
    ObjectStorageUploadInfo,
)
from saturnfs.schemas.usage import ObjectStorageUsageResults
from typing_extensions import Literal

DEFAULT_CALLBACK = NoOpCallback()


class SaturnFS(AbstractFileSystem):
    blocksize = settings.S3_MIN_PART_SIZE
    protocol = "sfs"

    def __init__(self):
        self.object_storage_client = ObjectStorageClient()
        self.file_transfer = FileTransferClient()
        weakref.finalize(self, self.close)
        super().__init__()

    @property
    def fsid(self) -> str:
        return "sfs_" + str(urlparse(settings.SATURN_BASE_URL).hostname)

    def get(  # pylint: disable=unused-argument
        self,
        rpath: Union[str, List[str]],
        lpath: Union[str, List[str]],
        recursive: Optional[bool] = None,
        callback: Callback = DEFAULT_CALLBACK,
        **kwargs,
    ):
        rpaths = self.expand_path(rpath, recursive=recursive)
        lpaths = make_path_posix(lpath)
        lpaths = other_paths(rpaths, lpaths)

        if len(rpaths) == 1:
            kwargs = {}
            callback.branch(rpaths[0], lpaths[0], kwargs)
            self.get_file(rpaths[0], lpaths[0], **kwargs)
        else:
            self.get_bulk(rpaths, lpaths, callback=callback)

    def put(
        self,
        lpath: Union[str, List[str]],
        rpath: Union[str, List[str]],
        recursive: bool = False,
        callback: Callback = DEFAULT_CALLBACK,
        **kwargs,
    ):
        rpath = self._strip_protocol(rpath)
        if isinstance(lpath, str):
            lpath = make_path_posix(lpath)

        fs = LocalFileSystem()
        lpaths = fs.expand_path(lpath, recursive=recursive)
        is_dir = isinstance(rpath, str) and self.isdir(rpath)

        # Overriding the default implementation to set exists=False
        # Without this, there's no way to format a recursive put
        # such that files always end up in the same location
        rpaths = other_paths(
            lpaths,
            rpath,
            exists=False,
            is_dir=is_dir,
        )

        callback.set_size(len(rpaths))
        for lp, rp in callback.wrap(zip(lpaths, rpaths)):
            callback.branch(lp, rp, kwargs)
            self.put_file(lp, rp, **kwargs)

    def copy(
        self,
        path1: Union[str, List[str]],
        path2: Union[str, List[str]],
        recursive: bool = False,
        on_error: Optional[str] = None,
        callback: Callback = DEFAULT_CALLBACK,
        maxdepth: Optional[int] = None,
        **kwargs,
    ):
        if on_error is None and recursive:
            on_error = "ignore"
        elif on_error is None:
            on_error = "raise"

        path1 = self.expand_path(path1, recursive=recursive, maxdepth=maxdepth)
        path2 = other_paths(path1, self._strip_protocol(path2))

        callback.set_size(len(path2))
        for p1, p2 in zip(path1, path2):
            callback.branch(p1, p2, kwargs)
            try:
                self.cp_file(p1, p2, **kwargs)
            except FileNotFoundError:
                if on_error == "raise":
                    raise

    def rm(
        self,
        path: Union[str, List[str]],
        recursive: bool = False,
        maxdepth: Optional[int] = None,
        callback: Callback = DEFAULT_CALLBACK,
    ):
        paths = self.expand_path(path, recursive=recursive, maxdepth=maxdepth)
        self.rm_bulk(paths, callback=callback)

    def mv(
        self,
        path1: Union[str, List[str]],
        path2: Union[str, List[str]],
        recursive: bool = False,
        maxdepth: Optional[int] = None,
        callback: Callback = DEFAULT_CALLBACK,
        rm_callback: Callback = DEFAULT_CALLBACK,
        **kwargs,
    ):
        self.copy(path1, path2, recursive=recursive, maxdepth=maxdepth, callback=callback, **kwargs)

        rm_kwargs: Dict[str, Callback] = {}
        if isinstance(path1, str):
            rm_callback.branch(path1, "", rm_kwargs)
        else:
            rm_callback.branch(f"{len(path1)} files", "", rm_kwargs)

        self.rm(path1, recursive=recursive, maxdepth=maxdepth, **rm_kwargs)

    @overload
    def ls(  # type: ignore[misc]
        self, path: str, detail: Literal[False] = False, **kwargs
    ) -> List[str]:
        # dummy code for pylint
        return []

    @overload
    def ls(self, path: str, detail: Literal[True] = True, **kwargs) -> List[ObjectStorageInfo]:
        # dummy code for pylint
        return []

    @overload
    def ls(
        self, path: str, detail: bool = False, **kwargs
    ) -> Union[List[str], List[ObjectStorageInfo]]:
        return []  # type: ignore[return-value]

    def ls(
        self,
        path: str,
        detail: bool = False,
        **kwargs,
    ) -> Union[List[str], List[ObjectStorageInfo]]:
        refresh = kwargs.get("refresh", False)
        path_is_dir = path.endswith("/")
        path = self._strip_protocol(path)

        results: Optional[List[ObjectStorageInfo]] = None
        if not refresh:
            results = self._ls_from_cache(path)

        if refresh or results is None:
            if "/" in path:
                # List object storage for an owner
                results = self._lsdir(path)
                if not results and not path_is_dir:
                    # Check for file exactly matching the given path
                    file_prefix = path.rsplit("/", 1)[-1]
                    results = self._lsdir(self._parent(path), file_prefix=file_prefix)
                    results = [info for info in results if not info.is_dir and info.name == path]
            elif path:
                # List owners with shared object storage in the org
                results = self._lsshared(path)
            else:
                results = self._lsorg()
            self.dircache[path] = results

        if detail:
            return results
        return sorted([info.name for info in results])

    def _ls_from_cache(self, path: str) -> Optional[List[ObjectStorageInfo]]:
        path = self._strip_protocol(path)
        if path in self.dircache:
            return self.dircache[path]

        parent = self._parent(path)
        parent_files: List[ObjectStorageInfo] = self.dircache.get(parent, None)
        if parent_files is not None:
            files = [
                f
                for f in parent_files
                if f.name == path or (f.name.rstrip("/") == path and f.is_dir)
            ]
            if len(files) == 0:
                # parent dir was listed but did not contain this file
                raise FileNotFoundError(path)
            elif len(files) == 1 and files[0].is_dir:
                # Cache contained the directory, but not its contents
                # List on a directory path should never return just the directory itself
                return None
            return files
        return None

    def _lsdir(self, dir: str, file_prefix: Optional[str] = None) -> List[ObjectStorageInfo]:
        """
        List contents of the given directory, optionally filtered by a file prefix
        """
        path = dir.rstrip("/") + "/"
        if file_prefix:
            path += file_prefix
        prefix = ObjectStoragePrefix.parse(path)

        files: List[ObjectStorageInfo] = []
        dirs: List[ObjectStorageInfo] = []
        for result in self.object_storage_client.list_iter(prefix):
            files.extend(result.files)
            dirs.extend(result.dirs)

        return dirs + files

    def _lsshared(self, org_name: str) -> List[ObjectStorageInfo]:
        """
        List owners that have shared object storage as "directories"
        """
        owners: List[ObjectStorageInfo] = []
        for result in self.object_storage_client.shared_iter(org_name):
            owners.extend(
                [
                    ObjectStorageInfo(
                        file_path="/",
                        owner_name=owner.name,
                        type="directory",
                    )
                    for owner in result.owners
                ]
            )
        return owners

    def _lsorg(self) -> List[ObjectStorageInfo]:
        """
        List orgs that the current identity has access to as "directories"
        """
        orgs = self.object_storage_client.orgs()
        return [
            ObjectStorageInfo(
                file_path="/",
                owner_name=org.name,
                type="directory",
            )
            for org in orgs
        ]

    @overload
    def find(  # type: ignore[misc]
        self,
        path: str,
        maxdepth: Optional[int] = None,
        withdirs: bool = False,
        detail: Literal[False] = False,
        **kwargs,
    ) -> List[str]:
        # dummy code for pylint
        return []

    @overload
    def find(
        self,
        path: str,
        maxdepth: Optional[int] = None,
        withdirs: bool = False,
        detail: Literal[True] = True,
        **kwargs,
    ) -> Dict[str, ObjectStorageInfo]:
        # dummy code for pylint
        return {}

    @overload
    def find(
        self,
        path: str,
        maxdepth: Optional[int] = None,
        withdirs: bool = False,
        detail: bool = False,
        **kwargs,
    ) -> Union[List[str], Dict[str, ObjectStorageInfo]]:
        # dummy code for pylint
        return []  # type: ignore[return-value]

    def find(
        self,
        path: str,
        maxdepth: Optional[int] = None,
        withdirs: bool = False,
        detail: bool = False,
        **kwargs,
    ) -> Union[List[str], Dict[str, ObjectStorageInfo]]:
        path = self._strip_protocol(path)
        if maxdepth is None and "/" in path:
            # Can list more efficiently by ignoring / delimiters rather than walking the file tree
            files: List[ObjectStorageInfo] = []
            prefix = ObjectStoragePrefix.parse(path + "/")

            for result in self.object_storage_client.list_iter(prefix, delimited=False):
                files.extend(result.files)
            if detail:
                return {file.name: file for file in files}
            return sorted(file.name for file in files)
        return super().find(path, maxdepth=maxdepth, withdirs=withdirs, detail=detail, **kwargs)

    @overload
    def glob(self, path: str, detail: Literal[False] = False, **kwargs) -> List[str]:  # type: ignore[misc]
        # dummy code for pylint
        return []

    @overload
    def glob(
        self, path: str, detail: Literal[True] = True, **kwargs
    ) -> Dict[str, ObjectStorageInfo]:
        # dummy code for pylint
        return {}

    @overload
    def glob(
        self, path: str, detail: bool = False, **kwargs
    ) -> Union[List[str], Dict[str, ObjectStorageInfo]]:
        # dummy code for pylint
        return []  # type: ignore[return-value]

    def glob(
        self, path: str, detail: bool = False, **kwargs
    ) -> Union[List[str], Dict[str, ObjectStorageInfo]]:
        return super().glob(path, detail=detail, **kwargs)

    @overload
    def walk(  # type: ignore[misc]
        self,
        path: str,
        maxdepth: Optional[int] = None,
        topdown: bool = True,
        detail: Literal[False] = False,
        **kwargs,
    ) -> Iterable[Tuple[str, List[str], List[str]]]:
        # dummy code for pylint
        yield "", [], []

    @overload
    def walk(
        self,
        path: str,
        maxdepth: Optional[int] = None,
        topdown: bool = True,
        detail: Literal[True] = True,
        **kwargs,
    ) -> Iterable[Tuple[str, Dict[str, ObjectStorageInfo], Dict[str, ObjectStorageInfo]]]:
        # dummy code for pylint
        yield "", {}, {}

    @overload
    def walk(
        self,
        path: str,
        maxdepth: Optional[int] = None,
        topdown: bool = True,
        detail: bool = False,
        **kwargs,
    ) -> Iterable[
        Union[
            Tuple[str, List[str], List[str]],
            Tuple[str, Dict[str, ObjectStorageInfo], Dict[str, ObjectStorageInfo]],
        ]
    ]:
        # dummy code for pylint
        yield "", [], []  # type: ignore[misc]

    def walk(
        self,
        path: str,
        maxdepth: Optional[int] = None,
        topdown: bool = True,
        detail: bool = False,
        **kwargs,
    ) -> Iterable[
        Union[
            Tuple[str, List[str], List[str]],
            Tuple[str, Dict[str, ObjectStorageInfo], Dict[str, ObjectStorageInfo]],
        ]
    ]:
        for root, dirs, files in super().walk(
            path, maxdepth=maxdepth, topdown=topdown, detail=detail, **kwargs
        ):
            yield root, dirs, files

    def exists(self, path: str, **kwargs) -> bool:
        if has_magic(path):
            # Avoid unecessary check in fsspec.
            # Paths with glob will never exist in saturn object storage
            return False
        try:
            self.info(path, **kwargs)
            return True
        except FileNotFoundError:
            return False

    def info(self, path: str, **kwargs) -> ObjectStorageInfo:
        path = self._strip_protocol(path)
        results = self.ls(self._parent(path), detail=True, **kwargs)
        for r in results:
            if r.name.rstrip("/") == path:
                return r

        # TODO: Don't think this is doing anything useful
        results = self.ls(path, detail=True, **kwargs)
        for r in results:
            if r.name.rstrip("/") == path:
                return r
        raise FileNotFoundError(path)

    @overload
    def open(
        self,
        path: str,
        mode: Union[Literal["rb"], Literal["wb"]] = "rb",
        block_size: Optional[int] = None,
        cache_options: Optional[Dict] = None,
        compression: Optional[str] = None,
        **kwargs,
    ) -> SaturnFile:
        # dummy code for pylint
        return SaturnFile(self, path)

    @overload
    def open(
        self,
        path: str,
        mode: Union[Literal["r"], Literal["w"]] = "r",
        block_size: Optional[int] = None,
        cache_options: Optional[Dict] = None,
        compression: Optional[str] = None,
        **kwargs,
    ) -> TextIOWrapper:
        # dummy code for pylint
        return TextIOWrapper(BytesIO())

    @overload
    def open(
        self,
        path: str,
        mode: str = "rb",
        block_size: Optional[int] = None,
        cache_options: Optional[Dict] = None,
        compression: Optional[str] = None,
        **kwargs,
    ) -> Union[TextIOWrapper, SaturnFile]:
        # dummy code for pylint
        return SaturnFile(self, path)

    def open(
        self,
        path: str,
        mode: str = "rb",
        block_size: Optional[int] = None,
        cache_options: Optional[Dict] = None,
        compression: Optional[str] = None,
        **kwargs,
    ) -> Union[TextIOWrapper, SaturnFile]:
        return super().open(
            path,
            mode=mode,
            block_size=block_size,
            cache_options=cache_options,
            compression=compression,
            **kwargs,
        )

    def _open(
        self,
        path: str,
        mode: str = "rb",
        block_size: Optional[int] = None,
        autocommit: bool = True,
        cache_options: Optional[Dict] = None,
        **kwargs,
    ) -> SaturnFile:
        return SaturnFile(
            self,
            path,
            mode=mode,
            block_size=block_size,
            autocommit=autocommit,
            cache_options=cache_options,
            **kwargs,
        )

    def put_file(
        self,
        lpath: str,
        rpath: str,
        callback: Callback = DEFAULT_CALLBACK,
        size: Optional[int] = None,
        block_size: Optional[int] = None,
        **kwargs,
    ):
        """Copy single file to remote"""
        if os.path.isdir(lpath):
            callback.set_size(0)
            callback.relative_update(0)
            return None

        with open(lpath, "rb") as f1:
            file_size = f1.seek(0, 2)
            if size is None:
                size = file_size
            elif file_size < size:
                raise SaturnError("File is smaller than the given size")
            callback.set_size(size)
            f1.seek(0)

            with self.open(rpath, mode="wb", block_size=block_size, size=size, **kwargs) as f2:
                while f1.tell() < size:
                    data = f1.read(self.blocksize)
                    segment_len = f2.write(data)
                    if segment_len is None:
                        segment_len = len(data)
                    callback.relative_update(segment_len)

            if size == 0:
                callback.relative_update(0)
        self.invalidate_cache(rpath)

    def cp_file(  # pylint: disable=unused-argument
        self,
        path1: str,
        path2: str,
        callback: Callback = DEFAULT_CALLBACK,
        block_size: Optional[int] = None,
        **kwargs,
    ):
        source = ObjectStorage.parse(path1)
        destination = ObjectStorage.parse(path2)

        try:
            presigned_copy = self.object_storage_client.start_upload(
                destination, part_size=block_size, copy_source=source
            )
        except SaturnError as e:
            if e.status == 404:
                raise FileNotFoundError(e.message) from e
            raise e

        size = sum(part.size for part in presigned_copy.parts)
        callback.set_size(size)

        done = False
        completed_parts: List[ObjectStorageCompletePart] = []
        while not done:
            try:
                for part in presigned_copy.parts:
                    completed_parts.append(self.file_transfer.copy_part(part))
                    callback.relative_update(part.size)
            except ExpiredSignature:
                # Get new presigned URLs
                next_part = len(completed_parts) + 1
                presigned_copy = self.object_storage_client.resume_upload(
                    presigned_copy.upload_id, first_part=next_part
                )
            else:
                done = True

        self.object_storage_client.complete_upload(presigned_copy.upload_id, completed_parts)
        self.invalidate_cache(path2)

    def get_file(
        self,
        rpath: str,
        lpath: str,
        callback: Callback = DEFAULT_CALLBACK,
        outfile: Optional[BytesIO] = None,
        **kwargs,
    ):
        super().get_file(rpath, lpath, callback=callback, outfile=outfile, **kwargs)

    def get_bulk(self, rpaths: List[str], lpaths: List[str], callback: Callback = DEFAULT_CALLBACK):
        callback.set_size(len(lpaths))
        downloads = self._iter_downloads(rpaths, lpaths)
        for download, lpath in callback.wrap(downloads):
            kwargs: Dict[str, Any] = {}
            callback.branch(download.name, lpath, kwargs)
            self.file_transfer.download(download, lpath, **kwargs)

    def _iter_downloads(
        self, rpaths: List[str], lpaths: List[str]
    ) -> Iterable[Tuple[ObjectStoragePresignedDownload, str]]:
        owner_downloads: Dict[str, Dict[str, str]] = {}
        for rpath, lpath in zip(rpaths, lpaths):
            remote = ObjectStorage.parse(rpath)
            owner_downloads.setdefault(remote.owner_name, {})
            owner_downloads[remote.owner_name][remote.file_path] = lpath

        for owner_name, downloads in owner_downloads.items():
            i = 0
            file_paths = list(downloads.keys())
            while i < len(file_paths):
                bulk = BulkObjectStorage(
                    owner_name=owner_name,
                    file_paths=file_paths[i : i + settings.OBJECT_STORAGE_MAX_LIST_COUNT],
                )
                bulk_download = self.object_storage_client.download_bulk(bulk)
                for download in bulk_download.files:
                    yield download, downloads[download.file_path]
                i += settings.OBJECT_STORAGE_MAX_LIST_COUNT

    def rm_file(self, path: str, callback: Callback = DEFAULT_CALLBACK):
        callback.set_size(1)
        remote = ObjectStorage.parse(path)
        self.object_storage_client.delete_file(remote)
        self.invalidate_cache(path)
        callback.relative_update(1)

    def rm_bulk(self, paths: List[str], callback: Callback = DEFAULT_CALLBACK):
        callback.set_size(len(paths))
        owner_paths: Dict[str, List[str]] = {}
        for path in paths:
            remote = ObjectStorage.parse(path)
            owner_paths.setdefault(remote.owner_name, [])
            owner_paths[remote.owner_name].append(remote.file_path)

        # Bulk delete by owner
        for owner_name, file_paths in owner_paths.items():
            # Delete in batches of 1000
            i = 0
            while i < len(file_paths):
                file_paths_chunk = file_paths[i : i + settings.OBJECT_STORAGE_MAX_LIST_COUNT]
                callback.relative_update(len(file_paths_chunk))
                bulk = BulkObjectStorage(
                    file_paths=file_paths[i : i + settings.OBJECT_STORAGE_MAX_LIST_COUNT],
                    owner_name=owner_name,
                )
                self.object_storage_client.delete_bulk(bulk)
                for path in bulk.file_paths:
                    self.invalidate_cache(path)
                i += settings.OBJECT_STORAGE_MAX_LIST_COUNT

    def list_uploads(
        self, path: str, is_copy: Optional[bool] = None
    ) -> List[ObjectStorageUploadInfo]:
        prefix = ObjectStoragePrefix.parse(path)
        return self.object_storage_client.list_uploads(prefix, is_copy=is_copy)

    def cancel_upload(self, upload_id: str):
        self.object_storage_client.cancel_upload(upload_id)

    def usage(self, owner_name: Optional[str] = None) -> ObjectStorageUsageResults:
        return self.object_storage_client.usage(owner_name)

    def created(self, path: str) -> datetime:
        info = self.info(path)
        if not info.is_dir:
            return info.created_at  # type: ignore[return-value]
        raise FileNotFoundError(path)

    def modified(self, path: str) -> datetime:
        info = self.info(path)
        if not info.is_dir:
            return info.updated_at  # type: ignore[return-value]
        raise FileNotFoundError(path)

    def close(self):
        self.object_storage_client.close()
        self.file_transfer.close()

    def invalidate_cache(self, path: Optional[str] = None):
        if path is None:
            self.dircache.clear()
        else:
            path = self._strip_protocol(path)
            self.dircache.pop(path, None)
            while path:
                self.dircache.pop(path, None)
                path = self._parent(path)

        super().invalidate_cache(path)


class SaturnFile(AbstractBufferedFile):
    """
    Open a remote object as a file. Data is buffered as needed.
    """

    fs: SaturnFS
    buffer: BytesIO
    blocksize: int
    offset: Optional[int] = None
    path: str

    def __init__(
        self,
        fs: SaturnFS,
        path: str,
        mode: str = "rb",
        block_size: Optional[int] = None,
        autocommit: bool = True,
        cache_type: str = "bytes",
        cache_options: Optional[Dict[str, Any]] = None,
        size: Optional[int] = None,
        **kwargs,
    ):
        if mode not in {"rb", "wb"}:
            raise NotImplementedError("File mode not supported")

        if block_size is None:
            if size is not None and size < settings.S3_MIN_PART_SIZE:
                block_size = size
            else:
                block_size = settings.S3_MIN_PART_SIZE
        elif block_size < settings.S3_MIN_PART_SIZE:
            raise SaturnError(f"Min block size: {settings.S3_MIN_PART_SIZE}")
        elif block_size > settings.S3_MAX_PART_SIZE:
            raise SaturnError(f"Max block size: {settings.S3_MIN_PART_SIZE}")

        self.remote = ObjectStorage.parse(path)
        self.size = size

        super().__init__(
            fs,
            path,
            mode,
            block_size,
            autocommit,
            cache_type=cache_type,
            cache_options=cache_options,
            size=size,
            **kwargs,
        )

        # Upload data
        self.upload_id: str = ""
        self.presigned_parts: List[ObjectStoragePresignedPart] = []
        self.completed_parts: List[ObjectStorageCompletePart] = []

        # Download data
        self.presigned_get: Optional[str] = None

    def _upload_chunk(self, final: bool = False) -> bool:
        num_bytes = self.buffer.tell()
        data: Optional[bytes] = None
        if not final and num_bytes < self.blocksize:
            # Defer upload until there are more blocks buffered
            return False
        else:
            self.buffer.seek(0)
            data = self.buffer.read(self.blocksize)

        if num_bytes > 0 or final and len(self.completed_parts) == 0:
            part_num = len(self.completed_parts) + 1
            self._presign_upload_parts(num_bytes, final=final)

            while data is not None:
                self._upload_part(data, part_num, num_bytes, final)

                # Get next chunk
                part_num += 1
                remaining = num_bytes - self.buffer.tell()
                if (final and remaining > 0) or (not final and remaining >= self.blocksize):
                    data = self.buffer.read(self.blocksize)
                elif remaining == 0:
                    data = None
                else:
                    # Defer upload of chunks smaller than
                    # blocksize until the last part
                    buffer = BytesIO()
                    buffer.write(self.buffer.read())
                    if self.offset is None:
                        self.offset = 0
                    self.offset += self.buffer.seek(0, 2)
                    self.buffer = buffer
                    return False

        if self.autocommit and final:
            self.commit()
        return not final

    def _upload_part(
        self, data: bytes, part_num: int, total_bytes: int, final: bool, retries: int = 5
    ):
        while retries > 0:
            part = self.presigned_parts[part_num - 1]
            try:
                completed_part = self.fs.file_transfer.upload_part(data, part)
            except ExpiredSignature:
                self._presign_upload_parts(total_bytes, final=final, force=True)
            else:
                self.completed_parts.append(completed_part)
                return

            retries -= 1

    def commit(self):
        if self.upload_id:
            self.fs.object_storage_client.complete_upload(self.upload_id, self.completed_parts)
            self.fs.invalidate_cache(self.path)

    def discard(self):
        if self.upload_id:
            self.fs.object_storage_client.cancel_upload(self.upload_id)

    def _initiate_upload(self):
        presigned_upload = self.fs.object_storage_client.start_upload(
            self.remote, self.size, self.blocksize
        )
        self.presigned_parts = presigned_upload.parts
        self.upload_id = presigned_upload.object_storage_upload_id

    def _fetch_range(self, start: int, end: int):
        if self.presigned_get is None:
            self._presign_download()

        headers = {"Range": f"bytes={start}-{end}"}
        try:
            response = self.fs.file_transfer.aws.get(self.presigned_get, headers=headers)
        except ExpiredSignature:
            self._presign_download()
            response = self.fs.file_transfer.aws.get(self.presigned_get, headers=headers)

        return response.content

    def _presign_download(self):
        self.presigned_get = self.fs.object_storage_client.download_file(self.remote).url

    def _presign_upload_parts(self, num_bytes: int, final: bool = False, force: bool = False):
        if self.blocksize > 0:
            num_parts = int(num_bytes / self.blocksize)
            remainder = num_bytes % self.blocksize
        else:
            num_parts = 1
            remainder = 0

        if final and remainder > 0:
            num_parts += 1

        num_completed = len(self.completed_parts)
        num_presigned = len(self.presigned_parts)
        last_part_size = remainder if final and remainder > 0 else None
        total_parts = num_completed + num_parts

        if force:
            # Throw out all unsued parts (expired)
            num_presigned = num_completed
            self.presigned_parts = self.presigned_parts[:num_presigned]
        elif final and num_presigned >= total_parts:
            if remainder > 0 and remainder != self.presigned_parts[total_parts - 1].size:
                # Fetch new final part with the correct size
                num_presigned = total_parts - 1
            else:
                num_presigned = total_parts

            # Throw out extra presigned parts
            self.presigned_parts = self.presigned_parts[:num_presigned]

        # Fetch new parts
        if num_presigned < total_parts:
            remaining_presigned = num_presigned - num_completed
            min_required = num_parts - remaining_presigned
            self._fetch_parts(
                num_completed + 1, min_required, final=final, last_part_size=last_part_size
            )

    def _fetch_parts(
        self,
        first_part: int,
        min_parts: int,
        final: bool = False,
        last_part_size: Optional[int] = None,
    ):
        """
        Get new presigned part URLs from saturn.

        Retrieves at least min_parts urls, but may fetch more to reduce overhead
        """
        if not final and min_parts < 10:
            num_parts = 10
        else:
            num_parts = min_parts

        retries = 5
        presigned_upload: Optional[ObjectStoragePresignedUpload] = None
        while presigned_upload is None and retries > 0:
            try:
                presigned_upload = self.fs.object_storage_client.resume_upload(
                    self.upload_id,
                    first_part,
                    first_part + num_parts - 1,
                    last_part_size=last_part_size,
                )
            except SaturnError as e:
                if final or retries == 0:
                    raise e
                elif e.status == 400:
                    # Check byte limit
                    usage = self.fs.usage(self.remote.owner_name)
                    remaining_bytes = usage.remaining_bytes
                    if remaining_bytes and remaining_bytes < num_parts * self.blocksize:
                        max_parts = int(remaining_bytes / self.blocksize)
                        if max_parts < min_parts:
                            raise e
                        num_parts = max_parts
            retries -= 1
        if presigned_upload is None:
            raise SaturnError("Failed to retrieve presigned upload")
        self.presigned_parts.extend(presigned_upload.parts)



from typing import Collection, Iterable, List, Optional

from requests import Session
from requests.adapters import HTTPAdapter, Retry
from saturnfs import settings
from saturnfs.api.delete import BulkDeleteAPI, DeleteAPI
from saturnfs.api.download import BulkDownloadAPI, DownloadAPI
from saturnfs.api.list import ListAPI, OrgListAPI, SharedAPI
from saturnfs.api.upload import UploadAPI
from saturnfs.api.usage import UsageAPI
from saturnfs.schemas.delete import ObjectStorageBulkDeleteResults
from saturnfs.schemas.download import (
    ObjectStorageBulkDownload,
    ObjectStoragePresignedDownload,
)
from saturnfs.schemas.list import (
    ObjectStorageListResult,
    ObjectStorageOrgs,
    ObjectStorageSharedResult,
    Org,
)
from saturnfs.schemas.reference import (
    BulkObjectStorage,
    ObjectStorage,
    ObjectStoragePrefix,
)
from saturnfs.schemas.upload import (
    ObjectStorageCompletedUpload,
    ObjectStorageCompletePart,
    ObjectStoragePresignedUpload,
    ObjectStorageUploadInfo,
    ObjectStorageUploadList,
)
from saturnfs.schemas.usage import ObjectStorageUsageResults


class ObjectStorageClient:
    """
    Manages session and and schemas for the ObjectStorage API
    """

    def __init__(
        self,
        retries: int = 10,
        backoff_factor: float = 0.1,
        retry_statuses: Collection[int] = frozenset([409, 423]),
    ):
        retry = Retry(retries, backoff_factor=backoff_factor, status_forcelist=retry_statuses)
        self.session = Session()
        self.session.headers["Authorization"] = f"token {settings.SATURN_TOKEN}"
        self.session.mount("http", HTTPAdapter(max_retries=retry))

    def start_upload(
        self,
        destination: ObjectStorage,
        size: Optional[int] = None,
        part_size: Optional[int] = None,
        copy_source: Optional[ObjectStorage] = None,
    ) -> ObjectStoragePresignedUpload:
        data = destination.dump()
        if size is not None:
            data["size"] = size
        if part_size:
            data["part_size"] = part_size
        if copy_source:
            data["copy_source"] = copy_source.dump()
        result = UploadAPI.start(self.session, data)
        return ObjectStoragePresignedUpload.load(result)

    def complete_upload(
        self, upload_id: str, completed_parts: List[ObjectStorageCompletePart]
    ) -> None:
        completed_upload = ObjectStorageCompletedUpload(parts=completed_parts)
        UploadAPI.complete(self.session, upload_id, completed_upload.dump())

    def cancel_upload(self, upload_id: str) -> None:
        UploadAPI.cancel(self.session, upload_id)

    def resume_upload(
        self,
        upload_id: str,
        first_part: Optional[int] = None,
        last_part: Optional[int] = None,
        last_part_size: Optional[int] = None,
    ) -> ObjectStoragePresignedUpload:
        result = UploadAPI.resume(self.session, upload_id, first_part, last_part, last_part_size)
        return ObjectStoragePresignedUpload.load(result)

    def list_uploads(
        self, prefix: ObjectStoragePrefix, is_copy: Optional[bool] = None
    ) -> List[ObjectStorageUploadInfo]:
        data = prefix.dump()
        if is_copy is not None:
            data["is_copy"] = is_copy
        result = UploadAPI.list(self.session, **data)
        return ObjectStorageUploadList.load_extended(result, prefix).uploads

    def delete_file(self, remote: ObjectStorage):
        DeleteAPI.delete(self.session, remote.dump())

    def delete_bulk(self, bulk: BulkObjectStorage):
        result = BulkDeleteAPI.delete(self.session, bulk.dump())
        return ObjectStorageBulkDeleteResults.load(result)

    def download_file(self, source: ObjectStorage) -> ObjectStoragePresignedDownload:
        result = DownloadAPI.get(self.session, source.dump())
        result.setdefault("owner_name", source.owner_name)
        return ObjectStoragePresignedDownload.load(result)

    def download_bulk(self, bulk: BulkObjectStorage) -> ObjectStorageBulkDownload:
        result = BulkDownloadAPI.get(self.session, bulk.dump())
        return ObjectStorageBulkDownload.load_extended(result, bulk.owner_name)

    def list(
        self,
        prefix: ObjectStoragePrefix,
        last_key: Optional[str] = None,
        max_keys: Optional[int] = None,
        delimited: bool = True,
    ) -> ObjectStorageListResult:
        result = ListAPI.get(
            self.session,
            **prefix.dump(),
            last_key=last_key,
            max_keys=max_keys,
            delimited=delimited,
        )
        return ObjectStorageListResult.load_extended(result, prefix=prefix)

    def list_iter(
        self, prefix: ObjectStoragePrefix, delimited: bool = True
    ) -> Iterable[ObjectStorageListResult]:
        last_key: Optional[str] = None
        while True:
            list_results = self.list(prefix, last_key, delimited=delimited)
            last_key = list_results.next_last_key
            yield list_results

            if not last_key:
                break

    def shared(
        self, org_name: str, last_key: Optional[str] = None, max_keys: Optional[int] = None
    ) -> ObjectStorageSharedResult:
        result = SharedAPI.get(
            self.session, org_name=org_name, last_key=last_key, max_keys=max_keys
        )
        return ObjectStorageSharedResult.load(result)

    def shared_iter(self, org_name: str) -> Iterable[ObjectStorageSharedResult]:
        last_key: Optional[str] = None
        while True:
            shared_results = self.shared(org_name, last_key=last_key)
            last_key = shared_results.next_last_key
            yield shared_results

            if not last_key:
                break

    def orgs(self) -> List[Org]:
        result = OrgListAPI.get(self.session)
        return [org for org in ObjectStorageOrgs.load(result).orgs if not org.locked]

    def usage(self, owner_name: Optional[str] = None) -> ObjectStorageUsageResults:
        result = UsageAPI.get(self.session, owner_name=owner_name)
        return ObjectStorageUsageResults.load(result)

    def close(self):
        self.session.close()



import os

from setuptools import find_packages, setup

install_requires = [
    "click",
    "requests",
    "marshmallow>=3.10.0,<4.0.0",
    "marshmallow-dataclass",
]


setup(
    name="saturnfs",
    version="0.0.1",
    maintainer="Saturn Cloud Developers",
    maintainer_email="dev@saturncloud.io",
    license="BSD-3-Clause",
    keywords="saturn cloud saturnfs object storage",
    description="Python library and CLI for interacting with object storage through saturn cloud",
    long_description=(open("README.md").read() if os.path.exists("README.md") else ""),
    long_description_content_type="text/markdown",
    entry_points={
        "console_scripts": [
            "saturnfs=saturnfs.cli.main:entrypoint",
        ],
    },
    url="https://saturncloud.io/",
    project_urls={
        "Documentation": "http://docs.saturncloud.io",
        "Source": "https://github.com/saturncloud/saturnfs",
        "Issue Tracker": "https://github.com/saturncloud/saturnfs/issues",
    },
    packages=find_packages(),
    install_requires=install_requires,
    zip_safe=False,
)



import json
from enum import Enum
from glob import has_magic
from typing import Any, Dict, List, Optional, Union

import click
from saturnfs.errors import SaturnError
from saturnfs.schemas.list import ObjectStorageInfo
from saturnfs.schemas.upload import ObjectStorageUploadInfo
from saturnfs.utils import human_readable_format


class OutputFormats(str, Enum):
    TABLE = "table"
    JSON = "json"

    def __str__(self) -> str:
        return str(self.value)

    @classmethod
    def values(cls) -> List[str]:
        return [str(value) for value in list(cls)]  # type: ignore

    @classmethod
    def validate(cls, format: str):
        if format not in cls.values():
            raise SaturnError(f'Unknown output format "{format}"')


def print_json(data: Union[List, Dict]):
    click.echo(json.dumps(data, indent=2))


def print_file_table(
    results: List[ObjectStorageInfo],
    human_readable: bool = False,
    relative_prefix: Optional[str] = None,
):
    headers = ["LastModified", "Size", "RelativePath" if relative_prefix else "Path"]
    data: List[List[str]] = []
    for info in results:
        last_modified = ""
        size = ""
        if info.updated_at:
            last_modified = info.updated_at.astimezone().strftime("%Y-%m-%d %H:%M:%S")
        if not info.is_dir:
            size = human_readable_format(info.size) if human_readable else str(info.size)
        if relative_prefix and info.name.startswith(relative_prefix):
            if info.name == relative_prefix:
                path = info.name.rsplit("/", 1)[-1]
            else:
                path = info.name[len(relative_prefix) :].lstrip("/")
        else:
            path = info.name
        data.append([last_modified, size, path])
    tabulate(data, headers, justify={"Size": ">"}, minwidth={"LastModified": 20, "Size": 10})


def print_upload_table(
    uploads: List[ObjectStorageUploadInfo], is_not_copy: bool = False, human_readable: bool = False
):
    headers = ["ID", "Expiration", "Size", "Name"]
    if not is_not_copy:
        headers.append("CopySource")

    data: List[List[str]] = []
    for upload in uploads:
        expires = upload.expires_at.astimezone().strftime("%Y-%m-%d %H:%M:%S")
        size = ""
        if upload.size is not None:
            size = human_readable_format(upload.size) if human_readable else str(upload.size)
        row = [upload.id, expires, size, upload.name]
        if not is_not_copy:
            row.append(upload.copy_source.name if upload.copy_source else "")
        data.append(row)

    tabulate(
        data, headers, justify={"Size": ">"}, minwidth={"ID": 32, "Expiration": 20, "Size": 10}
    )


def tabulate(
    data: List[List[Any]],
    headers: List[str],
    rpadding: int = 4,
    justify: Optional[Dict[str, str]] = None,
    minwidth: Optional[Dict[str, int]] = None,
):
    justify = justify if justify else {}
    minwidth = minwidth if minwidth else {}
    widths: List[int] = [0] * len(headers)
    for i, header in enumerate(headers):
        widths[i] = max(len(header), minwidth.get(header, 0))

    for row in data:
        for i, value in enumerate(row):
            widths[i] = max(widths[i], len(str(value)))

    header_format_str = ""
    format_str = ""
    for i, width in enumerate(widths):
        justify_char = justify.get(headers[i], "<")
        format_str += f"{{:{justify_char}{width}}}"
        header_format_str += f"{{:<{width}}}"
        if i < len(widths) - 1:
            format_str += " " * rpadding
            header_format_str += " " * rpadding

    click.echo(header_format_str.format(*headers))
    click.echo("-" * (sum(widths) + rpadding * (len(headers) - 1)))
    for row in data:
        click.echo(format_str.format(*row))


def strip_glob(path: str) -> str:
    """
    Return the longest full directory prefix of path with no glob characters

    e.g. org/identity/my/path*/tofiles?.txt -> org/identity/my/
    """
    if not has_magic(path):
        return path

    indstar = path.find("*") if path.find("*") >= 0 else len(path)
    indques = path.find("?") if path.find("?") >= 0 else len(path)
    indbrace = path.find("[") if path.find("[") >= 0 else len(path)
    first = min(indstar, indques, indbrace)

    path = path[:first]
    if "/" in path:
        return path.rsplit("/", 1)[0]
    return ""



from typing import Any, Dict, Optional

from requests import Session
from saturnfs.api.base import BaseAPI


class UploadAPI(BaseAPI):
    endpoint = "/api/object_storage/upload"

    @classmethod
    def start(cls, session: Session, data: Dict[str, Any]) -> Dict[str, Any]:
        url = cls.make_url()
        response = session.post(url, json=data)
        cls.check_error(response, 200)
        return response.json()

    @classmethod
    def complete(cls, session: Session, upload_id: str, data: Dict[str, Any]) -> None:
        url = cls.make_url(upload_id)
        response = session.post(url, json=data)
        cls.check_error(response, 204)

    @classmethod
    def cancel(cls, session: Session, upload_id: str) -> None:
        url = cls.make_url(upload_id)
        response = session.delete(url)
        cls.check_error(response, 204)

    @classmethod
    def resume(
        cls,
        session: Session,
        upload_id: str,
        first_part: Optional[int] = None,
        last_part: Optional[int] = None,
        last_part_size: Optional[int] = None,
    ) -> Dict[str, Any]:
        query_args = {
            "first_part": first_part,
            "last_part": last_part,
            "last_part_size": last_part_size,
        }
        url = cls.make_url(upload_id, query_args=query_args)
        response = session.get(url)
        cls.check_error(response, 200)
        return response.json()

    @classmethod
    def list(
        cls,
        session: Session,
        **query_args: Any,
    ) -> Dict[str, Any]:
        url = cls.make_url(query_args=query_args)
        response = session.get(url)
        cls.check_error(response, 200)
        return response.json()
